(* A pretty-printer for F. *)

(* We translate from System F to P and then print the P term (resp. type). *)

let new_tyvar =
  let id = ref 0 in
  fun () ->
    "a" ^ (string_of_int @@ Inferno.Utils.postincrement id)

let translate_tyvar env x =
  List.assoc x env

let rec translate_type (env : (F.tyvar * P.tyvar) list) (ty : F.nominal_type) : P.typ =
  let self = translate_type in
  match ty with
  | F.TyVar x ->
      P.TyVar (translate_tyvar env x)
  | F.TyArrow (ty1, ty2) ->
      P.TyArrow (self env ty1, self env ty2)
  | F.TyProduct tys ->
      P.TyProduct (List.map (self env) tys)
  | F.TySum tys ->
      P.TySum (List.map (self env) tys)
  | F.TyForall (x, ty) ->
      let alpha = new_tyvar () in
      P.TyForall (alpha, self ((x, alpha) :: env) ty)
  | F.TyMu (x, ty) ->
      let alpha = new_tyvar () in
      P.TyMu (alpha, self ((x, alpha) :: env) ty)
  | F.TyConstr (constr, tys) ->
      P.TyConstr (constr, List.map (self env) tys)

let print_type ty =
  Printer.print_type (translate_type [] ty)

let rec translate_term env (t : F.nominal_term) : P.term =
  let self = translate_term in
  match t with
  | F.Var x ->
      P.Var x
  | F.Hole (ty, ts) ->
      P.Hole (Some (translate_type env ty), List.map (self env) ts)
  | F.Annot (t, ty) ->
      P.Annot (self env t, ([], translate_type env ty))
  | F.Abs (x, ty, t) ->
      let annot = P.PAnnot (P.PVar x, ([], translate_type env ty)) in
      P.Abs (annot, self env t)
  | F.App (t1, t2) ->
      P.App (self env t1, self env t2)
  | F.Let (x, t, u) ->
      P.Let (P.PVar x, self env t, self env u)
  | F.TyAbs (x, t) ->
      let alpha = new_tyvar () in
      P.TyAbs (alpha, self ((x, alpha) :: env) t)
  | F.TyApp (t, ty) ->
      P.TyApp (self env t, translate_type env ty)
  | F.Tuple ts ->
      P.Tuple (List.map (self env) ts)
  | F.Proj (i, t) ->
      P.Proj (i, self env t)
  | F.Inj (tys, i, t) ->
      P.Inj (Some (List.map (translate_type env) tys), i, self env t)
  | F.LetProd (xs, t, u) ->
      let pat = P.PTuple (List.map (fun x -> P.PVar x) xs) in
      P.Let (pat , self env t, self env u)
  | F.Variant (lbl, (tid, tys) , t) ->
      P.Variant (
          lbl,
          Some (tid, List.map (translate_type env) tys),
          Some (self env t)
        )
  | F.Match (ty, t, brs) ->
      P.Match (
        Some (translate_type env ty),
        self env t,
        List.map (translate_branch env) brs
      )

and translate_branch env (pat, t) =
  (translate_pattern env pat, translate_term env t)

and translate_pattern env pat =
  let self = translate_pattern in
  match pat with
  | F.PVar x ->
      P.PVar x
  | F.PWildcard ->
      P.PWildcard
  | F.PAnnot (p, ty) ->
      P.PAnnot (self env p, ([], translate_type env ty))
  | F.PInj (tys, i, p) ->
      P.PInj (Some (List.map (translate_type env) tys), i, self env p)
  | F.PTuple ps ->
      P.PTuple (List.map (self env) ps)
  | F.PVariant (lbl, (tid, tys), p) ->
      P.PVariant (
          lbl,
          Some (tid, List.map (translate_type env) tys),
          Some (self env p)
        )

let print_term t =
  Printer.print_term (translate_term [] t)
