(* A pretty-printer for ML. *)

(* We translate from System F to P and then print the P term (resp. type). *)

let rec translate_type (ty : ML.typ) : P.typ =
  let self = translate_type in
  match ty with
  | ML.TyVar x ->
      P.TyVar x
  | ML.TyArrow (ty1, ty2) ->
      P.TyArrow (self ty1, self ty2)
  | ML.TyProduct tys ->
      P.TyProduct (List.map self tys)
  | ML.TyConstr (lbl, tys) ->
      P.TyConstr (lbl, List.map self tys)

let print_type (ty : ML.typ) =
  Printer.print_type (translate_type ty)

let translate_annot (vars, ty) =
  (vars, translate_type ty)

let rec translate_term (t : ML.term) : P.term =
  let self = translate_term in
  match t with
  | ML.Var x ->
      P.Var x
  | ML.Hole ts ->
      P.Hole (None, List.map self ts)
  | ML.Abs (x, t) ->
      P.Abs (P.PVar x, self t)
  | ML.App (t1, t2) ->
      P.App (self t1, self t2)
  | ML.Let (x, t, u) ->
      P.Let (P.PVar x, self t, self u)
  | ML.Annot (t, tyannot) ->
      P.Annot (self t, translate_annot tyannot)
  | ML.Tuple ts ->
      P.Tuple (List.map translate_term ts)
  | ML.LetProd (xs, t, u) ->
      let pat = P.PTuple (List.map (fun x -> P.PVar x) xs) in
      P.Let (pat, self t, self u)
  | ML.Variant (lbl, t) ->
      P.Variant (lbl, None, Option.map self t)
  | ML.Match (t, brs) ->
      P.Match (None, self t, List.map translate_branch brs)

and translate_branch (pat, t) =
  (translate_pattern pat, translate_term t)

and translate_pattern p =
  let self = translate_pattern in
  match p with
  | ML.PVar x ->
      P.PVar x
  | ML.PWildcard ->
      P.PWildcard
  | ML.PAnnot (p, tyannot) ->
      P.PAnnot (self p, translate_annot tyannot)
  | ML.PVariant (lbl, p) ->
      P.PVariant (lbl, None, Option.map self p)
  | ML.PTuple ps ->
      P.PTuple (List.map self ps)

let print_term t =
  Printer.print_term (translate_term t)

let to_string t =
  let b = Buffer.create 128 in
  PPrint.ToBuffer.pretty 0.9 80 b (print_term t);
  Buffer.contents b
